﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading;

namespace Callback
{
    class Program
    {
        static Func<Int32, BigInteger> factorize;

        static void Main(string[] args)
        {
            factorize = delegate(Int32 n)
            {
                BigInteger result = 1;
                for (Int32 i = 2; i < n; ++i)
                {
                    result *= i;
                    Thread.Sleep(100);
                }
                return result;
            };

            Console.WriteLine("[Sync] 20! = " + factorize.Invoke(20));
            Console.WriteLine("[Sync] 30! = " + factorize.Invoke(30));
            Console.WriteLine("[Sync] 40! = " + factorize.Invoke(40));

            factorize.BeginInvoke(20, EndFactorize, 20);
            factorize.BeginInvoke(25, EndFactorize, 20);
            factorize.BeginInvoke(30, EndFactorize, 30);
            factorize.BeginInvoke(35, EndFactorize, 35);
            factorize.BeginInvoke(40, EndFactorize, 40);

            Func<Int32, Int32> sum = delegate(Int32 n)
            {
                Int32 result = 1;
                for (Int32 i = 2; i < n; ++i)
                {
                    result += i;
                    Thread.Sleep(100);
                }
                return result;
            };
            
            sum.BeginInvoke(50,
                asyncState =>
                {
                    Console.WriteLine("[Async] sum(50) = {0}", sum.EndInvoke(asyncState));
                }, null);

            Console.ReadKey();
        }

        static void EndFactorize(IAsyncResult ar)
        {
            BigInteger result = factorize.EndInvoke(ar);
            Console.WriteLine("[Async] {0}! = {1}",  ar.AsyncState, result);
        }
    }
}
