﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;

namespace LINQ
{
    class FileEnumerable : IEnumerable<String>, IDisposable
    {
        private StreamReader _reader;

        public FileEnumerable(String path)
        {
            _reader = new StreamReader(path);
        }

        #region IEnumerable<T> interface

        public IEnumerator<String> GetEnumerator()
        {
            while (!_reader.EndOfStream)
                yield return _reader.ReadLine();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        #region IDisposable interface

        private Boolean _disposed = false;

        ~FileEnumerable()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);

            // Ne fusson le a destruktor majd...
            GC.SuppressFinalize(this); 
        }

        protected virtual void Dispose(Boolean disposing)
        {
            if (_disposed)
                return;

            if (disposing)
            {
                // Managed erőforrások felszabadítása
                _reader.Close();
            }

            // Unmanaged erőforrások felszabadítása
            // Most nincs ilyen.

            _disposed = true;
        }

        #endregion
    }
}
