﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;

namespace SyntaxSugar
{
    class Program
    {
        static void Main(string[] args)
        {

            #region Objektum inicializálások

            // Publikusan beállítható adattagok és tulajdonságok inicializálása konstruktorral.
            Person person = new Person("Kovács", "János", 25);

            // Bármelyik konstruktor is hívjuk, a publikusan beállítható adattagok és tulajdonságok külön inicializálhatóak.
            person = new Person { FamilyName = "Kiss", GivenName = "Gábor", Age = 37 };
            person = new Person { FamilyName = "Nagy", GivenName = "Ágnes" };
            person = new Person("Gipsz", "Júlia", 53) { Age = 42 };

            #endregion

            #region Gyűjtemény inicializálások

            // Tömbök inicializálása.
            Int32[] intArray = new Int32[] { 10, 20, 30 };

            // Implementálja az IEnumerable interfészt és legyen publikus Add(<érték>) metódusa. Az Add(<kulcs>, <érték>) művelet nem megfelelő.
            List<Int32> intList = new List<Int32> { 10, 20, 30 };
            RandomEnumerable<Int32> intRandomEnum = new RandomEnumerable<Int32> { 10, 20, 30 };

            #endregion

            #region Automatikus típusfelismerés

            // A statikus típus behelyettesítése fordítási idejű, és a nyelv erős típusosságát nem befolyásolja.
            var varA = 3; // ekkor var == Int32
            var varB = "hello"; // ekkor var == String
            var varC = new Int32[10]; // ekkor var == Int32[]

            Console.WriteLine("Automatikus típusfelismerés");
            Console.WriteLine("varA: " + varA.GetType().Name);
            Console.WriteLine("varB: " + varB.GetType().Name);
            Console.WriteLine("varC: " + varC.GetType().Name);
            Console.WriteLine();

            #endregion

            #region Anonim (névtelen) típusok

            /* 
             * A névtelen típusok (anonymous types) lehetővé teszik egyszerű,
             * rekordszerűobjektumok létrehozását, amelyeknek nem rendelkeznek önálló típussal.
             *   A névtelen típusra csak a varkulcsszóval hivatkozhatunk.
             *   A létrehozásukat objektum inicializálással kell elvégeznünk, tetszőleges tulajdonságneveket megadva.
             *   A megadott tulajdonságok publikusak, és csak olvashatóak lesznek.
             * 
             * A névtelen típusokhoz a fordítóprogram generál osztályt, ez közvetlenül az Object leszármazottja lesz.
             * 
             * A programkód szintjén nem elérhető az osztály, ezért 
             * közvetlenül nem manipulálhatjuk, nem készíthetünk hozzá metódusokat.
             * 
             * Két anononim típus pontosan akkor egyezik meg, h ugyanolyan statikus típusú
             * és nevű tulajdonságai vannak ugyanabban a sorrendben.
             */

            var anonimPerson1 = new { FamilyName = "Kovács", GivenName = "János", Age = 25 };
            var anonimPerson2 = new { FamilyName = "Kiss", GivenName = "Gábor", Age = 37 };
            var anonimPerson3 = new { FamilyName = "Nagy", GivenName = "Ágnes", Age = 19, Work = "Programozó" };
            var anonimPerson4 = new { Age = 42, FamilyName = "Gipsz", GivenName = "Júlia" };

            Console.WriteLine("Anonim (névtelen) típusok");
            Console.WriteLine("anonimPerson1: " + anonimPerson1.GetType().Name);
            Console.WriteLine("anonimPerson2: " + anonimPerson2.GetType().Name);
            Console.WriteLine("anonimPerson3: " + anonimPerson3.GetType().Name);
            Console.WriteLine("anonimPerson4: " + anonimPerson4.GetType().Name);
            Console.WriteLine();

            #endregion

            #region Bővítő metódusok

            String helloStr = "Hello World! ";
            String hello5Str = Extensions.Repeat(helloStr, 5);
            hello5Str = helloStr.Repeat(5);

            String name = Extensions.GetName(person);
            name = person.GetName();

            #endregion

            #region Lamda kifejezések

            /*
             * A lambda-kifejezések (lambda-expressions) funkcionális programozásból átvett elemek, amelyek egyszerre 
             * függvényként és objektumként is viselkednek. Lényegében a korábbi névtelen (delegált) műveletek helyett szolgálnak.
             * 
             * A lambda-kifejezést a => operátorral jelöljük, tőle balra a paraméterek, jobbra a művelet törzse írható le.
             * A kifejezést tetszőleges változóba elmenthetjük.
             * 
             * A lambda kifejezések típusa a sablonos Func<> típus, amely 1-17 sablontípust tud felvenni,
             * az utolsó a kifejezés eredménytípusa, előtte az argumentumok típusa szerepel.
             */

            Func<String, Boolean> less5 = delegate(String str)
            {
                return (str.Length < 5);
            };
            less5 = str => str.Length < 5;

            Func<Int32, Int32, Int32> add = delegate(Int32 a, Int32 b)
            {
                return a + b;
            };
            add = (a, b) => a + b;

            Func<Int32> const5 = delegate() { return 5; };
            const5 = () => 5;

            Func<String, String, Int32, Person> createPerson =
                delegate(String fn, String gn, Int32 a)
                {
                    return new Person(fn, gn, a);
                };
            createPerson = (fn, gn, a) => new Person(fn, gn, a);

            Func<String, String, Int32, Object> createNewType =
                delegate(String fn, String gn, Int32 a)
                {
                    return new { FamilyName = fn, GivenName = gn, Age = a };
                };
            createNewType = (fn, gn, a) => new { FamilyName = fn, GivenName = gn, Age = a };

            /*
             * Kifejezést futtatni atomi utasításként az Invoke() vagy az operator(),
             * párhuzamosított futtatás esetén a BeginInvoke() és EndInvoke() utasításokkal lehet.
             */
            Boolean result1 = less5.Invoke("hello");
            var result2 = add.Invoke(12, 34); // var == Int32
            var result3 = createPerson("Varga", "Elek", 28); // var == Person
            var result4 = createNewType("Varga", "Elek", 28); // var == Object

            /*
             * Func<T1, ..., Tn, void> == Action<T1, ..., Tn>
             * Func<T, Boolean> ~= Predicate<T>
             */

            #endregion

            #region Kifejezésfák

            /*
             * A kifejezésfák (expression trees)arra szolgálnak, hogy egy kifejezés szerkezetét.
             * Kifejezésfákat építhetünk futás közben a System.Linq.Expressions.Expression osztályon keresztül, 
             * amelyek kifejezésekké (akár lamda-kifejezésekké) alakíthatóak, így futás közben tudjuk legyártani,
             * illetve kiértékeltetni őket.
             * 
             * (Hasonlóan, mint ahogy a System.Reflectionnévtér osztályain keresztül lehetőségünk van
             * az osztályokhoz tartozó a metaobjektumokat manipulálni.)
             */

            Predicate<Int32> less10p = n => n < 10;

            // Változó a lambda-kifejezéshez
            var nPar = Expression.Parameter(typeof(Int32), "n");

            Expression<Predicate<Int32>> expression =
                Expression.Lambda<Predicate<Int32>>( // Lambda-kifejezés létrehozása 
                    Expression.LessThan(nPar, Expression.Constant(10)), // n < 10
                    nPar // Argumentumok
                );

            /*
             * A kész kifejezésfát a Compilemetódussal fordíthatjuk le,
             * az eredmény típusa az Expressionsablonjával egyezik meg.
             */

            Predicate<Int32> predicate = expression.Compile();
            var result5 = predicate(8); // var == Boolean

            #endregion

            Console.ReadKey();
        }
    }
}
